@@ -6,6 +6,8 @@ receiving response messages and subscribing for incoming messages from the devic
 
 import asyncio
 import logging
+import threading
+import traceback
 from concurrent.futures import ThreadPoolExecutor
 from threading import Thread
 from typing import Optional
@@ -14,6 +16,7 @@ from bleak import BleakClient
 from bleak import BleakScanner
 from bleak.backends.characteristic import BleakGATTCharacteristic
 from embodycodec import codec
+from embodycodec.exceptions import CrcError
 from embodyserial import embodyserial
 from embodyserial.helpers import EmbodySendHelper
 from packaging import version
@@ -54,21 +57,21 @@ class EmbodyBle(embodyserial.EmbodySender):
         response_msg_listener: Optional[ResponseMessageListener] = None,
     ) -> None:
         super().__init__()
-        self.__client: Optional[BleakClient] = None
-        self.__reader: Optional[_MessageReader] = None
-        self.__sender: Optional[_MessageSender] = None
-        self.__message_listeners: list[MessageListener] = []
+        self.__client: BleakClient | None = None
+        self.__reader: _MessageReader | None = None
+        self.__sender: _MessageSender | None = None
+        self.__message_listeners: set[MessageListener] = set()
+        self.__response_msg_listeners: set[ResponseMessageListener] = set()
+        self.__ble_message_listeners: set[BleMessageListener] = set()
+        self.__connection_listeners: set[ConnectionListener] = set()
         if msg_listener:
-            self.__message_listeners.append(msg_listener)
-        self.__response_msg_listeners: list[ResponseMessageListener] = []
+            self.__message_listeners.add(msg_listener)
         if response_msg_listener:
-            self.__response_msg_listeners.append(response_msg_listener)
-        self.__ble_message_listeners: list[BleMessageListener] = []
+            self.__response_msg_listeners.add(response_msg_listener)
         if ble_msg_listener:
-            self.__ble_message_listeners.append(ble_msg_listener)
-        self.__connection_listeners: list[ConnectionListener] = []
+            self.__ble_message_listeners.add(ble_msg_listener)
         if connection_listener:
-            self.__connection_listeners.append(connection_listener)
+            self.__connection_listeners.add(connection_listener)
         self.__connection_listener_executor = ThreadPoolExecutor(
             max_workers=1, thread_name_prefix="conn-worker"
         )
@@ -107,23 +110,23 @@ class EmbodyBle(embodyserial.EmbodySender):
             )
             or bool(d and d.name and d.name.lower() == self.__device_name.lower())
         )  # type: ignore[call-arg]
-        await scanner.stop()
         if not device:
             raise EmbodyBleError(
                 f"Could not find device with name {self.__device_name}"
             )
         self.__client = BleakClient(device, self._on_disconnected)
+        # self.__client._backend._mtu_size = 1497
         await self.__client.connect()
 
         logging.info(f"Connected: {self.__client}, mtu size: {self.__client.mtu_size}")
         self.__reader = _MessageReader(
-            self.__client, self.__message_listeners, self.__ble_message_listeners
+            self.__client,
+            self.__message_listeners,
+            self.__ble_message_listeners,
+            self.__response_msg_listeners,
         )
         self.__sender = _MessageSender(self.__client)
         self.__reader.add_response_message_listener(self.__sender)
-        if self.__response_msg_listeners:
-            for listener in self.__response_msg_listeners:
-                self.__reader.add_response_message_listener(listener)
         await self.__client.start_notify(
             UART_TX_CHAR_UUID, self.__reader.on_uart_tx_data
         )
@@ -144,6 +147,9 @@ class EmbodyBle(embodyserial.EmbodySender):
                 logging.debug(f"Failed to stop notify UART_TX:: {e}")
             await self.__client.disconnect()
             logging.info(f"Disconnected: {self.__client}")
+            if self.__reader:
+                self.__reader.stop()
+                self.__reader = None
             self.__client = None
 
     def shutdown(self) -> None:
@@ -232,7 +238,6 @@ class EmbodyBle(embodyserial.EmbodySender):
         await asyncio.sleep(timeout)
         await scanner.stop()
         devices = scanner.discovered_devices
-        await scanner.stop()
         return [
             d.name for d in devices if d.name and EmbodyBle.is_embody_ble_device(d.name)
         ]
@@ -245,13 +250,42 @@ class EmbodyBle(embodyserial.EmbodySender):
         )
 
     def add_message_listener(self, listener: MessageListener) -> None:
-        self.__message_listeners.append(listener)
+        self.__message_listeners.add(listener)
+        if self.__reader:
+            self.__reader.add_message_listener(listener)
+
+    def discard_message_listener(self, listener: MessageListener) -> None:
+        self.__message_listeners.discard(listener)
+        if self.__reader:
+            self.__reader.discard_message_listener(listener)
+
+    def add_response_message_listener(self, listener: ResponseMessageListener) -> None:
+        self.__response_msg_listeners.add(listener)
+        if self.__reader:
+            self.__reader.add_response_message_listener(listener)
+
+    def discard_response_message_listener(
+        self, listener: ResponseMessageListener
+    ) -> None:
+        self.__response_msg_listeners.discard(listener)
+        if self.__reader:
+            self.__reader.discard_response_message_listener(listener)
 
     def add_ble_message_listener(self, listener: BleMessageListener) -> None:
-        self.__ble_message_listeners.append(listener)
+        self.__ble_message_listeners.add(listener)
+        if self.__reader:
+            self.__reader.add_ble_message_listener(listener)
+
+    def discard_ble_message_listener(self, listener: BleMessageListener) -> None:
+        self.__ble_message_listeners.discard(listener)
+        if self.__reader:
+            self.__reader.discard_ble_message_listener(listener)
 
     def add_connection_listener(self, listener: ConnectionListener) -> None:
-        self.__connection_listeners.append(listener)
+        self.__connection_listeners.add(listener)
+
+    def discard_connection_listener(self, listener: ConnectionListener) -> None:
+        self.__connection_listeners.discard(listener)
 
     def __notify_connection_listeners(self, connected: bool) -> None:
         if len(self.__connection_listeners) == 0:
@@ -326,8 +360,9 @@ class _MessageReader:
     def __init__(
         self,
         client: BleakClient,
-        message_listeners: list[MessageListener],
-        ble_message_listeners: list[BleMessageListener],
+        message_listeners: Optional[set[MessageListener]] = None,
+        ble_message_listeners: Optional[set[BleMessageListener]] = None,
+        response_message_listeners: Optional[set[ResponseMessageListener]] = None,
     ) -> None:
         """Initialize MessageReader."""
         super().__init__()
@@ -338,9 +373,22 @@ class _MessageReader:
         self.__ble_message_listener_executor = ThreadPoolExecutor(
             max_workers=1, thread_name_prefix="ble-msg-worker"
         )
-        self.__message_listeners = message_listeners
-        self.__ble_message_listeners = ble_message_listeners
-        self.__response_message_listeners: list[ResponseMessageListener] = []
+        if message_listeners is not None:
+            self.__message_listeners = message_listeners
+        else:
+            self.__message_listeners = set()
+        self.__message_listeners_lock = threading.Lock()
+        if ble_message_listeners is not None:
+            self.__ble_message_listeners = ble_message_listeners
+        else:
+            self.__ble_message_listeners = set()
+        self.__ble_message_listeners_lock = threading.Lock()
+        if response_message_listeners is not None:
+            self.__response_message_listeners = response_message_listeners
+        else:
+            self.__response_message_listeners = set()
+        self.__response_message_listeners_lock = threading.Lock()
+        self.saved_data = bytearray()
 
     def stop(self) -> None:
         self.__message_listener_executor.shutdown(wait=False, cancel_futures=False)
@@ -366,15 +414,41 @@ class _MessageReader:
         New messages, both custom codec messages and BLE messages are received here.
         """
         logging.debug(f"New incoming data UART TX data: {bytes(data).hex()}")
-        try:
-            pos = 0
-            while pos < len(data):
+        if len(self.saved_data) > 0:
+            logging.debug(f"Adding saved data: {bytes(self.saved_data).hex()}")
+            data = self.saved_data + data
+            self.saved_data = bytearray()
+        pos = 0
+        while pos < len(data):
+            try:
                 msg = codec.decode(bytes(data[pos:]))
                 logging.debug(f"Decoded incoming UART message: {msg}")
                 self.__handle_incoming_message(msg)
                 pos += msg.length
-        except Exception as e:
-            logging.warning(f"Receive error during incoming message: {e}")
+            except BufferError as e:
+                # BufferError is sort of OK as we just need to aggregate the data for now?
+                logging.debug(
+                    f"Saving {len(data[pos:])} bytes for later parsing due to {e!r}."
+                )
+                self.saved_data = data[pos:]
+                break
+            except CrcError as e:
+                (msgtype, msglen) = codec.Message.get_meta(bytes(data[pos:]))
+                logging.warning(
+                    f"CRC error {e!r} at position {pos} for message type {hex(msgtype)} with length {msglen} in Data={data.hex()}"
+                )
+                pos += msglen  # Skip entire packet to resync, assuming fault is NOT in the length field!
+                continue
+            except Exception as e:
+                (msgtype, msglen) = codec.Message.get_meta(bytes(data[pos:]))
+                logging.warning(
+                    f"Receive error in on_uart_tx_data(): {e!r} at position {pos} of {len(data)} in Data={data.hex()}"
+                )
+                logging.warning(
+                    traceback.format_exception(Exception, e, e.__traceback__)
+                )
+                pos += msglen  # Skip message length to keep sync if message code was just unknown
+                continue
 
     def on_ble_message_received(
         self, uuid: BleakGATTCharacteristic, data: bytes
@@ -394,12 +468,11 @@ class _MessageReader:
 
     def __handle_message(self, msg: codec.Message) -> None:
         logging.debug(f"Handling new incoming message: {msg}")
-        if len(self.__message_listeners) == 0:
-            return
-        for listener in self.__message_listeners:
-            self.__message_listener_executor.submit(
-                _MessageReader.__notify_message_listener, listener, msg
-            )
+        with self.__message_listeners_lock:
+            for listener in self.__message_listeners:
+                self.__message_listener_executor.submit(
+                    _MessageReader.__notify_message_listener, listener, msg
+                )
 
     @staticmethod
     def __notify_message_listener(
@@ -410,15 +483,39 @@ class _MessageReader:
         except Exception as e:
             logging.warning(f"Error notifying listener: {str(e)}", exc_info=True)
 
+    def add_message_listener(self, listener: MessageListener) -> None:
+        with self.__message_listeners_lock:
+            self.__message_listeners.add(listener)
+
+    def discard_message_listener(self, listener: MessageListener) -> None:
+        with self.__message_listeners_lock:
+            self.__message_listeners.discard(listener)
+
     def add_response_message_listener(self, listener: ResponseMessageListener) -> None:
-        self.__response_message_listeners.append(listener)
+        with self.__response_message_listeners_lock:
+            self.__response_message_listeners.add(listener)
+        logging.warning(f"{listener!r} was added to response message listener set!")
+
+    def discard_response_message_listener(
+        self, listener: ResponseMessageListener
+    ) -> None:
+        with self.__response_message_listeners_lock:
+            self.__response_message_listeners.discard(listener)
+        logging.warning(f"{listener!r} was removed from response message listener set!")
+
+    def add_ble_message_listener(self, listener: BleMessageListener) -> None:
+        with self.__ble_message_listeners_lock:
+            self.__ble_message_listeners.add(listener)
+
+    def discard_ble_message_listener(self, listener: BleMessageListener) -> None:
+        with self.__ble_message_listeners_lock:
+            self.__ble_message_listeners.discard(listener)
 
     def __handle_response_message(self, msg: codec.Message) -> None:
         logging.debug(f"Handling new response message: {msg}")
-        if len(self.__response_message_listeners) == 0:
-            return
-        for listener in self.__response_message_listeners:
-            _MessageReader.__notify_rsp_message_listener(listener, msg)
+        with self.__response_message_listeners_lock:
+            for listener in self.__response_message_listeners:
+                _MessageReader.__notify_rsp_message_listener(listener, msg)
 
     @staticmethod
     def __notify_rsp_message_listener(
@@ -431,12 +528,11 @@ class _MessageReader:
 
     def __handle_ble_message(self, uuid: str, data: bytes) -> None:
         logging.debug(f"Handling new BLE message. UUID: {uuid}, data: {data.hex()}")
-        if len(self.__ble_message_listeners) == 0:
-            return
-        for listener in self.__ble_message_listeners:
-            self.__ble_message_listener_executor.submit(
-                _MessageReader.__notify_ble_message_listener, listener, uuid, data
-            )
+        with self.__ble_message_listeners_lock:
+            for listener in self.__ble_message_listeners:
+                self.__ble_message_listener_executor.submit(
+                    _MessageReader.__notify_ble_message_listener, listener, uuid, data
+                )
 
     @staticmethod
     def __notify_ble_message_listener(
@@ -446,6 +542,3 @@ class _MessageReader:
             listener.ble_message_received(uuid, data)
         except Exception as e:
             logging.warning(f"Error notifying ble listener: {str(e)}", exc_info=True)
-
-    def add_ble_message_listener(self, listener: BleMessageListener) -> None:
-        self.__ble_message_listeners.append(listener)
