@@ -147,6 +147,9 @@ class AttributeChangedListener:
     def on_autorec_changed(self, autorec: int) -> None:
         logging.info(f"Auto recording changed: {autorec}")
 
+    def on_flashinfo_changed(self, flashinfo: types.FlashInfo) -> None:
+        logging.info(f"Flash info changed: {flashinfo}")
+
 
 class AttributeChangedMessageListener(MessageListener, BleMessageListener):
     """MessageListener implementation delegating to high level callback interface."""
@@ -154,12 +157,12 @@ class AttributeChangedMessageListener(MessageListener, BleMessageListener):
     def __init__(
         self, attr_changed_listener: Optional[AttributeChangedListener] = None
     ) -> None:
-        self.__message_listeners: list[AttributeChangedListener] = []
+        self.__message_listeners: set[AttributeChangedListener] = set()
         if attr_changed_listener is not None:
             self.add_attr_changed_listener(attr_changed_listener)
 
     def add_attr_changed_listener(self, listener: AttributeChangedListener) -> None:
-        self.__message_listeners.append(listener)
+        self.__message_listeners.add(listener)
 
     def message_received(self, msg: codec.Message) -> None:
         """Process received message and delegate to listener callback."""
@@ -295,6 +298,9 @@ class AttributeChangedMessageListener(MessageListener, BleMessageListener):
             elif isinstance(msg.value, attributes.DisableAutoRecAttribute):
                 for listener in self.__message_listeners:
                     listener.on_autorec_changed(msg.value.value)
+            elif isinstance(msg.value, attributes.FlashInfoAttribute):
+                for listener in self.__message_listeners:
+                    listener.on_flashinfo_changed(msg.value.value)
             else:
                 logging.warning("Unhandled attribute changed message: %s", msg)
         elif isinstance(msg, codec.RawPulseChanged):
@@ -493,6 +499,14 @@ class EmbodyReporter:
     def stop_diagnostics_reporting(self) -> None:
         self.__send_reset_reporting(attributes.DiagnosticsAttribute.attribute_id)
 
+    def start_flash_reporting(self, int_millis: int) -> None:
+        self.__send_configure_reporting(
+            attributes.FlashInfoAttribute.attribute_id, int_millis
+        )
+
+    def stop_flash_reporting(self) -> None:
+        self.__send_reset_reporting(attributes.FlashInfoAttribute.attribute_id)
+
     def start_afe_settings_reporting(self, int_millis: int) -> None:
         self.__send_configure_reporting(
             attributes.AfeSettingsAllAttribute.attribute_id, int_millis
@@ -536,6 +550,7 @@ class EmbodyReporter:
         self.stop_recording_reporting()
         self.stop_sleep_mode_reporting()
         self.stop_temperature_reporting()
+        self.stop_flash_reporting()
 
     def __send_configure_reporting(
         self, attribute_id: int, interval: int, reporting_mode: int = 0x01
